import re
import xml.etree.ElementTree as ET
import matplotlib.pyplot as plt
import numpy as np
import matplotlib.lines as mpl

rgbpat = re.compile(r'rgb\((\d+),(\d+),(\d+)\)')

def fieldlines(svgfilename, ax=0):
    """plot fieldlines from svg file"""
    lines = []
    cols = []
    tree = ET.parse(svgfilename)
    root = tree.getroot()
    for line in root.findall('svg:line',
                             {'svg': 'http://www.w3.org/2000/svg'}):
        lines.append(
            ((float(line.attrib['x1']), float(line.attrib['x2'])),
             (float(line.attrib['y1']), float(line.attrib['y2']))))
        cols.append(
            [int(x)/256
             for x in rgbpat.findall(line.attrib['stroke'])[0]])
    a = np.array(lines)
    xmax, xmin = np.max(a[:, 0]), np.min(a[:, 0])
    ymax, ymin = np.max(a[:, 1]), np.min(a[:, 1])

    if ax == 0:
        ax = plt.gca()
    ax.set_frame_on(False)
    ax.set_aspect(1)
    ax.set_xlim((xmin, xmax))
    ax.set_ylim((ymax, ymin))  # upside down
    ax.set_yticks([])
    ax.set_xticks([])

    for l, c in zip(lines, cols):
        ax.add_line(mpl.Line2D(l[0], l[1], color=c))



"""zmq functions for FEMAG

"""
import re
import pathlib
import threading
import json
import time
import logging
try:
    import zmq
except ImportError:
    pass

numpat = re.compile(r'([+-]?\d+(?:\.\d+)?(?:[eE][+-]\d+)?)\s*')
logger = logging.getLogger(__name__)

class ProtFile:
    def __init__(self, dirname, num_cur_steps):
        self.size = 0
        self.looplen = 0
        self.cur_steps = [1, num_cur_steps]
        self.n = 0
        self.num_loops = 0
        import platform
        self.dirname = dirname
        self.name = 'samples'

    def percent(self):
        if self.looplen > 0:
            return min(100 * self.n / self.looplen, 100)
        return 0

    def update(self):
        if not self.dirname:
            return ''
        p = list(pathlib.Path(self.dirname).glob('*.PROT'))
        if p:
            buf = ''
            if self.size < p[0].stat().st_size:
                with p[0].open() as fp:
                    fp.seek(self.size)
                    buf = fp.read()
            return self.append(buf)
        return ''

    def append(self, buf):
        self.size += len(buf)
        for line in [l.strip() for l in buf.split('\n') if l]:
            if line.startswith('Loop'):
                self.n = 0
                try:
                    cur_steps = self.cur_steps[self.num_loops]
                except IndexError:
                    cur_steps = 1
                x0, x1, dx, nbeta = [float(f)
                                     for f in re.findall(numpat, line)][:4]
                move_steps = round((x1-x0)/dx+1)
                beta_steps = int(nbeta)
                self.looplen = cur_steps*beta_steps*move_steps
                self.num_loops += 1
            elif (line.startswith('Cur') or
                  line.startswith('Id')):
                self.n += 1
            elif line.startswith('Number movesteps Fe-Losses'):
                return f'{self.percent():3.1f}%' # 100%
            elif line.startswith('begin'):
                self.name = line.split()[1].strip()

        return f'{self.percent():3.1f}%'  # {self.n}/{self.looplen}'


class SubscriberTask(threading.Thread):
    ylabel_index = 1
    curve_label = '{}.'
    # used by static notify func
    percent_list = []
    notify_timerfunc = None
    notify_send_loop = True
    notify = None
    notify_send_header = set()
    notify_send_data = dict()
    # progress, xydata and this entries of this list
    # will send only only once per timestep
    simple_data = ['license']

    def __init__(self, **kwargs):
        threading.Thread.__init__(self)
        context = zmq.Context.instance()
        self.subscriber = context.socket(zmq.SUB)
        self.port = kwargs.get('port', None)
        self.host = kwargs.get('host')
        self.notify = kwargs.get('notify', None)
        SubscriberTask.notify = kwargs.get('notify', None)
        self.header = kwargs.get('header')
        self.num_cur_steps = kwargs.get('num_cur_steps', None)
        SubscriberTask.curve_label = kwargs.get('curve_label', '')
        SubscriberTask.timestep = kwargs.get('timestep', 2)

        if not self.host:
            self.host = 'localhost'
        if not self.header:
            self.header = [b'']
        self.running = True

        # timer function
        if not SubscriberTask.notify_timerfunc:
            SubscriberTask.notify_timerfunc = threading.Timer(0.1, SubscriberTask.send_notify)
            SubscriberTask.notify_send_loop = True
            SubscriberTask.notify_timerfunc.start()

        if b'xyplot' in self.header:
            self.ylabel = self.curve_label.format(SubscriberTask.ylabel_index)
            SubscriberTask.ylabel_index += 1
        if b'progress' in self.header:
            self.protfile = ProtFile(None, self.num_cur_steps)
            self.protId = len(SubscriberTask.percent_list)
            SubscriberTask.percent_list.append(0)  # 0%

        self.subscriber.connect(f'tcp://{self.host}:{self.port}')
        self.subscriber.setsockopt(zmq.SUBSCRIBE, self.header[0] if len(self.header) == 1 else b'')
        self.controller = zmq.Context.instance().socket(zmq.PULL)
        self.controller_url = f'inproc://publisher{self.port}'
        try:
            self.controller.bind(self.controller_url)
        except zmq.error.ZMQError:
            pass # ignore

        self.poller = zmq.Poller()
        self.poller.register(self.subscriber, zmq.POLLIN)
        self.poller.register(self.controller, zmq.POLLIN)
        self.logger = logger

    def stop(self):
        socket = zmq.Context.instance().socket(zmq.PUSH)
        socket.connect(self.controller_url)
        socket.send(b"quit")
        socket.close()
        self.running = False

    def clear():
        SubscriberTask.ylabel_index = 1
        SubscriberTask.curve_label = '{}.'
        SubscriberTask.notify_timerfunc = None
        SubscriberTask.notify_send_loop = False
        SubscriberTask.notify = None
        SubscriberTask.notify_send_header = set()
        SubscriberTask.notify_send_data = dict()
        SubscriberTask.percent_list = []

    def send_notify():
        logger.debug(f"Send loop: {SubscriberTask.notify_send_loop}, timestep: {SubscriberTask.timestep}")
        while SubscriberTask.notify_send_loop:
            if 'progress_logger' in SubscriberTask.notify_send_header:
                # collect data from different threads
                SubscriberTask.notify_send_header.remove('progress_logger')
                numTot = len(SubscriberTask.percent_list)
                d = json.loads(SubscriberTask.notify_send_data.get('progress_logger')[1])
                d['percent'] = sum(SubscriberTask.percent_list) / numTot
                d['subtitle'] = f"{SubscriberTask.percent_list.count(100)} of {numTot}" if numTot > 1 else ''
                SubscriberTask.notify(['progress_logger', json.dumps(d)])
            if 'xyplot' in SubscriberTask.notify_send_header:
                SubscriberTask.notify([s.decode('latin1')
                                       for s in SubscriberTask.notify_send_data.get('xyplot')])
                SubscriberTask.notify_send_header.remove('xyplot')

            # simple
            for sdata in SubscriberTask.simple_data:
                if sdata in SubscriberTask.notify_send_header:
                    SubscriberTask.notify([s.decode('latin1')
                                           for s in SubscriberTask.notify_send_data.get(sdata)])
                    SubscriberTask.notify_send_header.remove(sdata)

            time.sleep(abs(SubscriberTask.timestep))
        logger.debug(f"Send Finished loop: {SubscriberTask.notify_send_loop}")

    def run(self):
        self.logger.debug("subscriber is ready, port: %s", {self.port})
        while self.running:
            socks = dict(self.poller.poll())
            if socks.get(self.subscriber) == zmq.POLLIN:
                try:
                    response = self.subscriber.recv_multipart()
                    # Sometimes femag send messages with only len = 1. These messages must be ignored
                    if len(response) < 2:
                        continue
                    # header progress
                    if response[0] == b'progress' and b'progress' in self.header:
                        SubscriberTask.notify_send_header.add('progress_logger')
                        response[0] = b'progress_logger'
                        SubscriberTask.notify_send_data['progress_logger'] = response
                        while self.protId >= len(SubscriberTask.percent_list):
                            # keep old progressbar working
                            SubscriberTask.percent_list.append(0)
                        SubscriberTask.percent_list[self.protId] = json.loads(response[1].decode()).get('percent')
                        continue

                    # header xyplot (add ylabel)
                    if response[0] == b'xyplot' and b'xyplot' in self.header    :
                        d = json.loads(response[1].decode(), strict=False)
                        d['ylabel'] = f"{d.get('ylabel')}_{self.ylabel}" \
                            if d.get('ylabel') else self.ylabel
                        response[1] = json.dumps(d).encode()

                        # timestep negative, immediately update
                        if SubscriberTask.timestep < 0:
                            self.notify([s.decode('latin1') for s in response])
                        else:
                            SubscriberTask.notify_send_data['xyplot'] = response
                            SubscriberTask.notify_send_header.add('xyplot')
                        continue

                    # simple
                    for sdata in SubscriberTask.simple_data:
                        if response[0] == sdata.encode() and sdata.encode() in self.header:
                            SubscriberTask.notify_send_header.add(sdata)
                            SubscriberTask.notify_send_data[sdata] = response
                            continue

                    if response[0] not in self.header:
                        self.notify([s.decode('latin1') for s in response])

                except Exception:
                    self.logger.error(
                        "error in subscription message processing", exc_info=True)

            if socks.get(self.controller) == zmq.POLLIN:
                req = self.controller.recv()
                self.logger.debug("subscriber %s", req)
                break
        self.subscriber.close()
        self.controller.close()
        self.logger.debug("subscriber stopped")



# -*- coding: utf-8 -*-
"""Managing machine and simulation model parameters

"""
import logging
import string
import numpy as np
from . import windings

logger = logging.getLogger(__name__)
#
# Set of legal model name chars
# 0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ
MODELNAME_CHARS = string.printable[:63] + "-_äöüéè"
# maximum name length
MAX_MODEL_NAME_LEN = 255

default_th_properties = {
    "spmaweight": 8.96,
    "tempcoef": 3.9e-3,
    "thcond": 30,
    "thcap": 480
}

def movesteps(nodes):
    """returns list of move steps

    Args:
       nodes: number of nodes in airgap
    """
    return [nodes // w + 1 for w in range(1, nodes//2) if nodes % w == 0]


class MCerror(Exception):
    pass


class Model(object):
    def __init__(self, parameters):
        if isinstance(parameters, dict):
            for k in parameters.keys():
                setattr(self, k, parameters[k])

    def set_value(self, name, value, p=None):
        """set value of parameter identified by name

        Args:
            name: name of parameter
            value: value to be assigned to parameter
        """
        if isinstance(name, str):
            setattr(self, name, value)
            return

        if len(name) > 1:
            k = name[0]
            if hasattr(self, k):
                self.set_value(name[1:], value, getattr(self, k))
                return
            elif p:
                self.set_value(name[1:], value, p[k])
                return
            self.set_value(name[1:], value, self)
            return
        if p:
            p[name[0]] = value
            return
        setattr(self, name[0], value)

    def get(self, name, r=None):
        """return value of key name

        Args:
            name (str or list of str): key of parameter value

        Return:
            value of parameter identified by key
        """
        try:
            if isinstance(name, str):
                if r is not None:
                    return getattr(self, name, r)
                return getattr(self, name)
            if r and type(r) == dict:
                for k in name:
                    r = r.get(k)
                return r
            if len(name) > 1:
                if r:
                    return self.get(name[1:], getattr(r, name[0]))
                return self.get(name[1:], getattr(self, name[0]))
            return getattr(self, name[0])
        except KeyError as e:
            logger.error(e)
            raise MCerror(e)

    def __getitem__(self, name):
        return getattr(self, name)

    def __str__(self):
        "return string format of this object"
        return repr(self.__dict__)

    def __repr__(self):
        "representation of this object"
        return self.__str__()


class MachineModel(Model):
    """represents a machine model for a FE analysis

    Args:
      parameters: string or dict containing the model parameters. For example:
    ::

        {'lfe': 0.1,
        'poles': 4,
        'outer_diam': 0.13,
        'bore_diam': 0.07,
        'stator':{
        'num_slots': 12,
        'num_slots_gen': 3,
        ...
        },
        'magnet':{
            'material': 'M395',
            ..
        }
        }

        if parameters is string it is interpreted as the model name.
    """

    def __init__(self, parameters):
        super(self.__class__, self).__init__(parameters)
        name = 'DRAFT'
        if isinstance(parameters, str):
            name = parameters
            self.connect_full = False  # no matter
        else:
            if 'name' in parameters:
                name = parameters['name']
            if 'windings' in parameters:
                self.winding = self.windings

            # connect model even for complete model (see fsl connect_models)
            self.connect_full = parameters.get('afmtype', '') == ''
        # must sanitize name to prevent femag complaints
        self.name = ''.join([n
                             for n in name.strip()
                             if n in MODELNAME_CHARS][:MAX_MODEL_NAME_LEN])
        try:
            self.external_rotor = (self.external_rotor == 1)
        except AttributeError:
            self.external_rotor = False
        self.move_inside = 1.0 if self.external_rotor else 0.0
        if 'exit_on_end' not in parameters:
            self.exit_on_end = 'false'
        if 'magnet' in parameters:
            if 'mcvkey_mshaft' in self.magnet:
                self.magnet['mcvkey_shaft'] = self.magnet['mcvkey_mshaft']
            for mcv in ('mcvkey_yoke', 'mcvkey_shaft'):
                if mcv not in self.magnet:
                    self.magnet[mcv] = 'dummy'
        if 'coord_system' in parameters or 'afmtype' in parameters:
            self.move_action = 1
            wdg = windings.Winding({'Q': self.stator['num_slots'],
                                    'p': self.poles//2,
                                    'm': self.winding.get('num_phases', 3),
                                    'l': self.winding.get('num_layers', 1)})
            self.winding['wdgscheme'] = ''.join([
                '{'] + [','.join([''.join(['{']+[','.join([''.join([
                    '{', ','.join(
                        [str(n) for n in z]), '}']) for z in l])] + ['}'])
                                  for l in wdg.zoneplan()])] + ['}'])
            if not hasattr(self, 'pole_width'):  # note: for preview usage only
                self.pole_width = np.pi * self.outer_diam/self.poles
        else:
            self.coord_system = 0
            self.move_action = 0

        try:
            self.set_num_slots_gen()
        except (AttributeError):
            pass
        try:
            if 'wdgtype' not in self.winding:
                self.winding['wdgtype'] = 'SYM'
        except AttributeError:
            pass

        try:
            self.commutator = self.winding['wdgtype'] == 'CMM'
        except AttributeError:
            self.commutator = False
        try:
            self.winding['cufilfact'] = self.winding['fillfac']
        except (KeyError, AttributeError):
            pass

    def slot_height(self):
        if 'statorRotor3' in self.stator:
            return self.stator['statorRotor3']['slot_height']
        if 'stator1' in self.stator:
            return self.stator['stator1']['slot_rf1'] - self.stator['stator1']['tip_rh1']
        if 'stator4' in self.stator:
            return self.stator['stator4']['slot_height']
        da1 = self.bore_diam
        if self.external_rotor:
            yh = da1-self.inner_diam
        else:
            yh = da1+self.outer_diam
        return 0.6*yh/2

    def slot_area(self):
        if 'slot_area' in self.stator:
            return self.stator['slot_area']
        da1 = self.bore_diam
        Q1 = self.stator['num_slots']
        hs = self.slot_height()
        return np.pi*hs*(da1+hs)/Q1/2

    def set_num_slots_gen(self):
        if 'num_slots_gen' not in self.stator:
            try:
                m = self.winding['num_phases']
            except (KeyError, AttributeError):
                m = 1

            if hasattr(self, 'magnet'):
                slotgen = [m*self.poles]
            else:
                slotgen = [self.poles]
            try:
                slotgen.append(int(self.stator['num_slots']))
            except KeyError:
                pass

            try:
                slotgen.append(int(self.rotor['num_slots']))
            except (AttributeError, KeyError):
                pass
            logging.debug(slotgen)
            g = np.gcd.reduce(slotgen)
            if hasattr(self, 'magnet') and g > 1:
                g //= m
            if 'num_slots' in self.stator:
                Q1 = self.stator['num_slots']
                self.stator['num_slots_gen'] = Q1 // g
            else:
                try:
                    Q2 = self.rotor['num_slots']
                    self.rotor['num_slots_gen'] = Q2 // g
                except (AttributeError, KeyError):
                    pass

    def set_mcvkey_magnet(self, mcvkey):
        self.mcvkey_magnet = mcvkey

    def get_mcvkey_magnet(self):
        try:
            return self.mcvkey_magnet
        except AttributeError:
            return ''

    def _set_magnet_material(self, comp, magcurves, magnetmat):
        names = []
        missing = []
        magnet = 0
        mcv = 0
        try:
            if magnetmat:
                magnet = magnetmat.find(comp['material'])
            if magnet and 'mcvkey' in magnet:
                if magcurves:
                    mcv = magcurves.find(magnet['mcvkey'])
                if mcv:
                    logger.debug('magnet mcv %s', mcv)
                    comp['mcvkey_magnet'] = mcv
                    names.append((mcv, 1.0))
                else:
                    missing.append(magnet['mcvkey'])
                    logger.error('magnet mcv %s not found',
                                 magnet['mcvkey'])
        except KeyError:
            pass
        except AttributeError:
            if 'material' in comp:
                missing.append(comp['material'])
                logger.error('magnet material %s not found',
                             comp['material'])
        return names, missing

    def set_magcurves(self, magcurves, magnetmat={}):
        """set and return real names of magnetizing curve material

        Args:
            magcurves: :class: 'MagnetizingCurve' including
                              magnetizing curve materials

        Return:
            set of magnetizing curve names with fillfac that have to be created

        """
        names = []
        missing = []
        thkeys = 'thcap', 'thcond'
        mcv = 0
        if 'stator' in self.__dict__:
            fillfac = self.stator.get('fillfac', 1.0)
            for k in ('mcvkey_yoke', 'mcvkey_teeth'):
                mcvname = k+'_name'
                if mcvname not in self.stator:
                    try:
                        if self.stator[k] != 'dummy':
                            if magcurves:
                                mcv = magcurves.find(self.stator[k])
                            if mcv:
                                logger.debug('stator mcv %s', mcv)
                                self.stator[k] = magcurves.fix_name(
                                    mcv, fillfac)
                                names.append((mcv, fillfac))
                                self.stator[mcvname] = mcv
                                m = magcurves.find_by_name(mcv)
                                if m and set(thkeys).issubset(m.keys()):
                                    for k in thkeys:
                                        self.stator[k] = m[k]
                                    logger.info('stator mcv %s set therm prop in stator',
                                                mcv)
                            else:
                                missing.append(self.stator[k])
                                logger.error('stator mcv %s not found',
                                             self.stator[k])
                    except KeyError:
                        pass

                else:  # mcvname in self.stator:
                    names.append((self.stator[mcvname], fillfac))
            n, m = self._set_magnet_material(self.stator, magcurves, magnetmat)
            names += n
            missing += m

        if 'magnet' in self.__dict__:
            rotor = self.magnet
            subregion = 'magnet'
        elif 'rotor' in self.__dict__:
            rotor = self.rotor
            subregion = 'rotor'
        else:
            # no magnet no rotor
            return set(names)

        fillfac = rotor.get('fillfac', 1.0)
        try:
            if (rotor['mcvkey_yoke'] != 'dummy' and
                    'mcvkey_yoke_name' not in rotor):
                if magcurves:
                    mcv = magcurves.find(rotor['mcvkey_yoke'])
                if mcv:
                    logger.debug('%s mcv %s',
                                 subregion, mcv)
                    rotor['mcvkey_yoke'] = magcurves.fix_name(
                        mcv, fillfac)
                    rotor['mcvkey_yoke_name'] = mcv
                    m = magcurves.find_by_name(mcv)
                    if m and set(thkeys).issubset(m.keys()):
                        for k in thkeys:
                            rotor[k] = m[k]
                        logger.debug('stator mcv %s set therm prop in rotor', mcv)
                    names.append((mcv, fillfac))
                else:
                    missing.append(rotor['mcvkey_yoke'])
                    logger.error('%s mcv %s not found',
                                 subregion,
                                 rotor['mcvkey_yoke'])
            elif 'mcvkey_yoke_name' in rotor:
                names.append((rotor['mcvkey_yoke_name'], fillfac))

        except KeyError:
            pass

        try:
            if (rotor['mcvkey_shaft'] != 'dummy' and
                    'mcvkey_shaft_name' not in rotor):
                if magcurves:
                    mcv = magcurves.find(rotor['mcvkey_shaft'])
                if mcv:
                    logger.debug('shaft mcv %s', mcv)
                    rotor['mcvkey_shaft'] = magcurves.fix_name(mcv)
                    mshaft = magcurves.find_by_name(mcv)
                    thkeys_shaft = ['thcond', 'thcap', 'rho']
                    if mshaft and set(thkeys_shaft).issubset(mshaft.keys()):
                        for k in thkeys_shaft:
                            rotor[k+'_shaft'] = mshaft[k]
                        rotor['spmaweight_shaft'] = rotor['rho_shaft']

                    self.stator['mcvkey_shaft_name'] = mcv
                    names.append((mcv, 1.0))
                else:
                    missing.append(rotor['mcvkey_shaft'])
                    logger.error(' shaft %s not found',
                                 rotor['mcvkey_shaft'])
            elif 'mcvkey_shaft_name' in rotor:
                names.append((self.stator['mcvkey_shaft_name'], 1.0))

        except KeyError:
            pass

        if 'magnet' in self.__dict__:
            n, m = self._set_magnet_material(self.magnet, magcurves, magnetmat)
            missing += m
            names += n
        if missing:
            raise MCerror("Material properties missing: {}".format(
                ', '.join(set(missing))))
        return set(names)

    def statortype(self):
        """return type of stator slot"""
        for k in self.stator:
            if isinstance(self.stator[k], dict):
                return k
        raise AttributeError("Missing stator slot model in {}".format(
            self.stator))

    def rotortype(self):
        """return type of rotor slot"""
        for k in self.rotor:
            # anything that is a dict must represent the type
            if isinstance(self.rotor[k], dict):
                if k == 'rot_hsm':
                    return 'EESM'
                return k
        raise AttributeError("Missing rotor model in {}".format(self.magnet))

    def magnettype(self):
        """return type of magnet slot"""
        for k in self.magnet:
            if k not in {'material', 'temp_prop'}:
                if isinstance(self.magnet[k], dict):
                    return k
        raise AttributeError("Missing magnet model in {}".format(self.magnet))

    def is_complete(self):
        """check completeness of models"""
        if self.is_dxffile() or self.is_svgfile():
            return True
        try:
            self.statortype()
            if hasattr(self, 'rotor'):
                self.rotortype()
            else:
                self.magnettype()
            return True
        except AttributeError:
            return False

    def is_dxffile(self):
        if 'dxffile' in self.__dict__:
            if isinstance(self.dxffile, dict):
                return True
        return False

    def is_svgfile(self):
        if 'svgfile' in self.__dict__:
            if isinstance(self.svgfile, dict):
                return True
        return False

    def has_magnet(self):  # either rotor or magnet
        return hasattr(self, 'magnet')

    def props(self):
        """returns dict of this model"""
        keys = ('name', 'poles', 'outer_diam', 'airgap',
                'bore_diam', 'inner_diam', 'external_rotor', 'agndst')
        model = {k: getattr(self, k) for k in keys if hasattr(self, k)}
        if hasattr(self, 'stator'):
            model['stator'] = {k: self.stator[k]
                               for k in ('num_slots', 'num_slots_gen', 'slot_area')
                               if k in self.stator}
        return model

class FeaModel(Model):
    """represents a simulation model for a FE analysis"""
    def __init__(self, parameters):
        self.recsin = ''  # recalc mcv for dynamic simulation
        self.cufilfact = 0.45
        self.culength = 1.4
        self.wind_temp = 20
        self.slot_indul = 0.0
        self.skew_angle = 0.0
        self.num_skew_steps = 0
        self.num_par_wdgs = 1
        self.eval_force = 0
        self.optim_i_up = 0
        self.plots = []
        self.airgap_induc = []
        super(self.__class__, self).__init__(parameters)
        if parameters.get('calculationMode', '') == 'asyn_motor':
            self.recsin = 'flux'

    def get_num_cur_steps(self):
        """returns number of curSteps (used for progress calc)"""
        try:
            if self.calculationMode == 'psd_psq_fast':
                return round(
                    ((self.maxiq-self.minid)/self.delta_id) + 1)
            if self.calculationMode == 'ld_lq_fast':
                return (self.num_cur_steps+1)  # must include 0
        except AttributeError as e:
            logger.warning("%s current steps of %s",
                           e, self.calculationMode)
            pass
        return 1



import pathlib
import logging
import json
import numpy as np
import scipy.optimize as so
from scipy.interpolate import make_interp_spline
import scipy.integrate as ig
import femagtools.parstudy

logger = logging.getLogger('shortcircuit')

def _parstudy_list(femag, result_func):
    workdir = femag.workdir
    magnetMat = femag.magnets
    magnetizingCurves = femag.magnetizingCurves
    condMat = femag.condMat
    templatedirs = femag.templatedirs
    cmd = femag.cmd
    return femagtools.parstudy.List(
        workdir, condMat=condMat, magnets=magnetMat,
        magnetizingCurves=magnetizingCurves,
        cmd=cmd, result_func=result_func)

def get_shortCircuit_parameters(bch, nload):
    """extracts shortciruit parameters from bch"""
    try:
        if nload < 0:
            nload = 0
        if nload > 2:
            nload = 2
        if nload > 0:
            dqld = bch.dqPar['ld']
            dqlq = bch.dqPar['lq']
            dqpsim = bch.dqPar['psim']
            if len(dqld) <= nload or len(dqlq) <= nload or len(dqpsim) <= nload:
                ld = dqld[-1]/bch.armatureLength
                lq = dqlq[-1]/bch.armatureLength
                psim = dqpsim[-1]/bch.armatureLength
            else:
                ld = dqld[nload-1]/bch.armatureLength
                lq = dqlq[nload-1]/bch.armatureLength
                psim = dqpsim[nload-1]/bch.armatureLength
        else:
            ld = bch.machine['ld']/bch.armatureLength
            lq = bch.machine['lq']/bch.armatureLength
            psim = bch.machine['psim']/bch.armatureLength
        return dict(
            r1=bch.machine['r1'],
            ld=ld,
            lq=lq,
            Hk=bch.magnet['demag_hx'],
            Tmag=bch.magnet['Tmag'],
            psim=psim,
            num_pol_pair=bch.machine['p'],
            fc_radius=bch.machine['fc_radius'],
            lfe=bch.armatureLength/1e3,
            pocfilename=bch.machine['pocfile'],
            num_par_wdgs=bch.machine.get('num_par_wdgs', 1),
            calculationMode='shortcircuit')
    except (KeyError, AttributeError, IndexError):
        raise ValueError("missing pm/Rel-Sim results")


def find_peaks_and_valleys(t, y):
    """ return peak and valley of y with maximum amplitude
    """
    peaks = (np.diff(np.sign(np.diff(y))) < 0).nonzero()[0] + 1
    if len(peaks>0):
        ip = np.argmax(y[peaks])
        pv = {'ip': y[peaks][ip], 'tp': t[peaks][ip]}
    else:
        pv = {'ip': [], 'tp': []}
    valleys = (np.diff(np.sign(np.diff(y))) > 0).nonzero()[0] + 1
    if len(valleys>0):
        iv = np.argmin(y[valleys])
        pv.update({'iv': y[valleys][iv], 'tv': t[valleys][iv]})
    else:
        pv.update({'iv': [], 'tv': []})
    pv.update({'peaks': y[peaks], 'valleys': y[valleys],
               'tpeaks': t[peaks], 'tvalleys': t[valleys]})
    return pv

def shortcircuit(femag, machine, bch, simulation, engine=0):
    scdata = {}
    calcmode = simulation.get('calculationMode', '')
    simulation.update(
        get_shortCircuit_parameters(bch,
                                    simulation.get('initial', 2)))
    if 'speed' not in simulation:
        simulation['speed'] = bch.dqPar['speed']
    if simulation.get('sc_type', 3) == 3:
        logger.info("3phase short circuit simulation")
        builder = femagtools.fsl.Builder(femag.templatedirs)
        fslcmds = (builder.open_model(femag.model) +
                   builder.create_shortcircuit(simulation))
#                   ['save_model("cont")'])
        fslfile = 'shortcircuit.fsl'
        (pathlib.Path(femag.workdir)/fslfile).write_text(
            '\n'.join(fslcmds),
            encoding='latin1', errors='ignore')
        femag.run(fslfile)  # options?
        bchfile = femag.get_bch_file(femag.modelname)
        if bchfile:
            bchsc = femagtools.bch.Reader()
            logger.info("Read BCH %s", bchfile)
            bchsc.read(pathlib.Path(bchfile).read_text(
                encoding='latin1', errors='ignore'))
            bchsc.scData['demag'] = bchsc.demag
            if simulation.get('sim_demagn', 0):
                dd = {'displ': [d['displ']
                                for d in bchsc.demag if 'displ' in d],
                      'H_max': [d['H_max']
                                for d in bchsc.demag if 'H_max' in d],
                      'H_av': [d['H_av']
                               for d in bchsc.demag if 'H_av' in d]}
                x1 = bchsc.demag[0]['current_1']
                x2 = bchsc.demag[0]['current_2']
                def func(phi):
                    return x2*np.cos(phi) - x1*np.cos(phi-2*np.pi/3)
                phi=so.fsolve(func, 0)[0]
                i1max = x1/np.cos(phi)

                phirot = dd['displ'][0]/180*np.pi
                logger.info("i1max %g phi %g phirot %g", i1max, phi, phirot)
                bchsc.scData['demag'] = demag(
                    femag, machine, simulation,
                    i1max, phirot, phi, engine)
                bchsc.scData['demag'].update(dd)
            scdata = bchsc.scData
            #for w in bch.flux:
            #    try:
            #        bch.flux[w] += bchsc.flux[w]
            #        bch.flux_fft[w] += bchsc.flux_fft[w]
            #    except (KeyError, IndexError):
            #        logging.debug(
            #            "No additional flux data in sc simulation")
            #        break

    if simulation.get('sc_type', 3) == 2:
        if 'i1max' not in simulation:
            # just a wild guess
            simulation['i1max'] = 5*bch.machine['i1']
        logger.info("2phase short circuit simulation i1max = %.0f",
                    simulation['i1max'])
        if 'magn_temp' not in simulation:
            simulation['magn_temp'] = bch.magnet.get('Tmag', 20)
        scdata = shortcircuit_2phase(femag, machine, simulation, engine)

    else:
        logger.warning("Empty shortcircuit results for type %s",
                       simulation.get('sc_type', 'unknown'))
    # must reset calcmode
    if calcmode:
        simulation['calculationMode'] = calcmode
    else:
        del simulation['calculationMode']
    return scdata

def sc_result_func(task):
    basedir = pathlib.Path(task.directory)
    psitorq = np.loadtxt(basedir/'psi-torq-rot.dat')
    pos = np.unique(psitorq[:, 0])
    ncurs = psitorq[:, 0].shape[0]//pos.shape[0]
    ire = psitorq[:ncurs, 1:4]
    psire = np.reshape(psitorq[:, 4:7], (-1, ncurs, 3))
    torq = np.reshape(psitorq[:, 7], (-1, ncurs))
    return {'pos': pos.tolist(), 'ire': ire.tolist(),
            'psire': psire.tolist(), 'torq': torq.tolist()}


def shortcircuit_2phase(femag, machine, simulation, engine=0):
    i1max = simulation['i1max']
    num_cur_steps = 4
    i1 = np.linspace(0, i1max, num_cur_steps)
    i1vec = np.concat((-i1[::-1], i1[1:]))
    flux_sim = {
        'calculationMode': 'psi-torq-rot',
        'i1max': i1max,
        'curvec': [],
        'magntemp': simulation['magn_temp'],
        'fc_radius': simulation['fc_radius'],
        'num_par_wdgs': simulation['num_par_wdgs']}

    if engine:
        parstudy = _parstudy_list(femag, sc_result_func)
        parvardef = {
            "decision_vars": [
                {"values": i1vec, "name": "curvec"}]
        }
        results = parstudy(parvardef, machine, flux_sim, engine)

        ire = np.array([r['ire'][0] for r in results['f']])
        pos = np.array(results['f'][0]['pos'])
        phi = pos*np.pi/180
        torq = np.hstack([r['torq'] for r in results['f']])
        psire = np.hstack([r['psire'] for r in results['f']])
    else:
        simulation.update(flux_sim)
        simulation['curvec'] = i1vec.tolist()
        results = femag(machine, simulation, fslfile='2ph_shortcircuit.fsl')
        class Task:
            def __init__(self, workdir):
                self.directory = workdir
        results = sc_result_func(Task(femag.workdir))
        ire = np.array(results['ire'])
        pos = np.array(results['pos'])
        torq = np.array(results['torq'])
        psire = np.array(results['psire'])

    #with open('results.json', 'w') as fp:
    #    json.dump({'ire': ire.tolist(), 'pos': pos.tolist(),
    #               'torq': torq.tolist(), 'psire': psire.tolist()}, fp)
    logger.info("move steps %d currents %s", len(pos), ire[:,0])

    Ai = [femagtools.utils.fft(pos, psire[:, k, 0])['a']
          for k in range(np.shape(psire)[1])]
    A = make_interp_spline(ire[:,0], Ai)
    A0i = [femagtools.utils.fft(pos, psire[:, k, 0])['a0']
           for k in range(np.shape(psire)[1])]
    A0 = make_interp_spline(ire[:,0], A0i)
    Bi = [femagtools.utils.fft(pos, psire[:, k, 1])['a']
          for k in range(np.shape(psire)[1]-1, -1, -1)]
    B = make_interp_spline(ire[::-1,1], Bi)
    B0i = [femagtools.utils.fft(pos, psire[:, k, 1])['a0']
           for k in range(np.shape(psire)[1]-1, -1, -1)]
    B0 = make_interp_spline(ire[::-1,1], B0i)
    alfa0_ai = [femagtools.utils.fft(pos, psire[:, k, 0])['alfa0']
                for k in range(np.shape(psire)[1])]
    alfa0_a = make_interp_spline(ire[:,0], alfa0_ai)
    alfa0_bi = [femagtools.utils.fft(pos, psire[:, k, 1])['alfa0']
                for k in range(np.shape(psire)[1]-1, -1, -1)]
    alfa0_b = make_interp_spline(ire[::-1,1], alfa0_bi)

    Tqi = [femagtools.utils.fft(pos, torq[:, k])['a']
           for k in range(np.shape(torq)[1])]
    Tq = make_interp_spline(ire[:, 0], Tqi)
    Tq0i = [femagtools.utils.fft(pos, torq[:, k])['a0']
            for k in range(np.shape(torq)[1])]
    Tq0 = make_interp_spline(ire[:, 0], Tq0i)
    alfa0_t = [femagtools.utils.fft(pos, torq[:, k])['alfa0']
               for k in range(np.shape(torq)[1])]

    T0 = np.mean([femagtools.utils.fft(pos, psire[:, k, 0])['T0']
                  for k in range(np.shape(psire)[1])])
    pp = 360/T0

    def torque(phi, i):
        try:
            alfa0 = np.ones(len(i))*np.mean(alfa0_t)
            alfa0[i < 0] = alfa0_t[0]
            alfa0[i > 0] = alfa0_t[-1]
        except TypeError:
            alfa0 = np.mean(alfa0_t)
            if i < 0:
                alfa0 = alfa0_t[0]
            if i > 0:
                alfa0 = alfa0_t[-1]
        return Tq(i)*np.cos(pp*phi+alfa0) + Tq0(i)

    def psia(phi, i):
        return A(i)*np.cos(pp*phi+alfa0_a(i))+A0(i)

    def psib(phi, i):
        return B(i)*np.cos(pp*phi+alfa0_b(i))+B0(i)

    def dpsiadi(phi,i):
        return A(i, nu=1)*np.cos(pp*phi+alfa0_a(i))+A0(i,nu=1)
    def dpsiadphi(phi,i):
        return -pp*A(i)*np.sin(pp*phi+alfa0_a(i))
    def dpsibdi(phi,i):
        return B(i, nu=1)*np.cos(pp*phi+alfa0_b(i))+B0(i,nu=1)
    def dpsibdphi(phi,i):
        return -pp*B(i)*np.sin(pp*phi+alfa0_b(i))

    speed = simulation['speed']
    r1 = simulation['r1']
    l1s = simulation.get('l1s',0)
    wm = 2*np.pi*speed
    w1 = pp*wm

    def didt(t, y):
        return [((2*r1*y[0] + wm*(
            dpsiadphi(y[1],y[0]) - dpsibdphi(y[1],-y[0])))/
            (-dpsiadi(y[1],y[0]) - dpsibdi(y[1],-y[0]) -2*l1s)),
                wm]
    tmin = simulation.get('tstart', 0)
    tmax = simulation.get('simultime', 0.1)
    nsamples = simulation.get('nsamples', 400)
    t = np.linspace(tmin, tmax, nsamples)

    def func(x):
        return B(0)*np.sin(pp*x+alfa0_b(0)) - A(0)*np.sin(pp*x+alfa0_a(0))
    phi0 = so.fsolve(func, [0])[0]

    Y0 = [0, phi0]
    sol = ig.solve_ivp(didt, (t[0], t[-1]), Y0, dense_output=True)
    ia = sol.sol(t).T[:, 0]
    pv = find_peaks_and_valleys(t, ia)
    iap = pv['tp'], pv['ip']
    iav = pv['tv'], pv['iv']
    iac = pv['tpeaks'][-1], pv['peaks'][-1]

    logger.info("Ia %.1f %.1f %.1f (phi0 %.4f)",
                iap[1], iav[1], iac[1], phi0)

    def func(x):
        y = torque(wm*t+phi0+x, ia)
        pv = find_peaks_and_valleys(t, y)
        return pv['peaks'][-1] + pv['valleys'][-1]

    dphi = so.fsolve(func, [0])[0]
    torque = torque(wm*t+phi0+dphi, ia)
    pv = find_peaks_and_valleys(t, torque)
    tp = pv['tp'], pv['ip']
    tv = pv['tv'], pv['iv']
    tc = pv['tpeaks'][-1], pv['peaks'][-1]
    logger.info("Torque %.1f %.1f %.1f (dphi %.4f)",
                tp[1], tv[1], tc[1], dphi)

    # rotor position at maximum current:
    trot = min(iav[0], iap[0])
    phirot = wm*trot + phi0
    logger.debug("phirot %.1f", phirot)

    scData = {
        'ia': ia.tolist(),
        'ib': (-ia).tolist(),
        'ic': np.zeros(ia.shape).tolist(),
        'time': t.tolist(),
        'torque': torque.tolist(),
        'speed': speed,
        'ikd': iac[1],
        'tkd': tc[1],
        'iks': iap[1] if iap[1] > abs(iav[1]) else iav[1],
        'tks': tp[1] if tp[1] > abs(tv[1]) else tv[1]
    }
    scData['peakWindingCurrents'] = [float(scData['iks']),
                                     -float(scData['iks']), 0]
    if simulation.get('sim_demagn', 0):
        i1max = iap[1] if iap[1] > abs(iav[1]) else iav[1]
        scData['demag'] = demag(femag, machine, simulation,
                                i1max, phirot, 0, engine)
    return scData

def demag(femag, machine, simulation, i1max, phirot, phi, engine=0):
    """demag simulation using psi-torq-rem-rot"""
    logger.info("Demagnetization processing")
    i1min = simulation.get('i1min', abs(i1max/3))
    num_steps = simulation.get('num_demag_cur_steps', 7)
    b = (i1min-abs(i1max))/np.log(i1min/abs(i1max))
    a = abs(i1max)/b
    xtab = np.linspace(i1min/abs(i1max),
                       1+2*(1-i1min/abs(i1max))/(num_steps-1), num_steps+2)
    i1tab = b*(a+np.log(xtab))

    if simulation.get('sc_type', 3) == 3:
        curvec = [[a*np.cos(phi), a*np.cos(phi-2*np.pi/3),
                   a*np.cos(phi+2*np.pi/3)] for a in i1tab]
    else:
        if i1max > 0:
            curvec = [[a, -a, 0] for a in i1tab]
        else:
            curvec = [[-a, a, 0] for a in i1tab]
    simulation.update({
        'calculationMode': 'psi-torq-rem',
        'phi': phirot,
        'magntemp': simulation['Tmag'],
        'curvec': curvec})
    try:
        del simulation['airgap_induc']
    except KeyError:
        pass
    _ = femag(machine, simulation, fslfile='demag.fsl')

    ptr = np.loadtxt(pathlib.Path(femag.workdir) / "psi-torq-rem.dat")
    i1 = np.concat(([0], np.max(ptr[:,1:4], axis=1)))
    rr = np.concat(([1], ptr[:,-1]))
    dmag = {'Hk': simulation['Hk'],
            'Tmag': simulation['Tmag'],
            'i1': i1.tolist(),
            'i1max': float(np.abs(i1max)),
            'rr': rr.tolist()}
    # critical current
    if np.min(rr) < 0.99:
        k = np.where(np.diff(dmag['rr'])<-1e-3)[0][0]+1
        dmag['i1c'] = float(i1[k])
        if abs(i1max) < i1[-1]:
            rrf = make_interp_spline(i1, rr)
            dmag['rr_i1max'] = float(rrf(abs(i1max)))
    return dmag



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
import pytest
import femagtools.mcv
import os
import sys
import tempfile
import shutil


mcvPars = [dict(
    name="TKS_NO_20",
    desc="PowerCore NO 20 ;ThyssenKrupp Steel Eur",
    cversion=0,
    ctype=1,
    recalc=1,
    curve=[{"bi": [0.0, 0.09029103, 0.17855673, 0.26691416, 0.35827994],
            "bi2": [11, 22, 33],
            "nuer": [111, 222, 333],
            "a": [6, 5, 4],
            "b": [9, 8, 7],
            "hi": [1, 2, 3, 4, 5]}],
    remz=0.0,
    fillfac=0.92000002,
    bsat=0.0,
    bref=0.0,
    ch=4.0,
    ch_freq=5.0,
    cw=3.0,
    cw_freq=2.0,
    fo=50.0,
    Bo=1.5,
    b_coeff=1.0,
    rho=7.6500001,
    fe_sat_mag=2.15)
]


def test_findById():
    mcv = femagtools.mcv.MagnetizingCurve(mcvPars)
    result = mcv.find('TKS_NO_20')
    expected = mcvPars[0]['name']
    assert result == expected


def test_writeFile():
    dir = tempfile.mkdtemp()
    ext = '.MC' if sys.platform == 'win32' else '.MCV'
    mcv = femagtools.mcv.MagnetizingCurve(mcvPars)
    result = mcv.writefile('TKS_NO_20', dir)
    assert result == mcvPars[0]['name'] + ext

    mcv = femagtools.mcv.read(os.path.join(dir, result))
            
    assert pytest.approx(
        mcvPars[0]['curve'][0]['bi']) == mcv.get_results()['curve'][0]['bi']
    assert pytest.approx(
        mcvPars[0]['curve'][0]['hi']) == mcv.get_results()['curve'][0]['hi']

    shutil.rmtree(dir)


def test_writeFile_fillfac():
    ext = '.MC' if sys.platform == 'win32' else '.MCV'
    testPath = os.path.split(__file__)[0]
    mcv = femagtools.mcv.read(os.path.join(testPath, 'data/TKS_NO_20.MCV'))
    dir = tempfile.mkdtemp()
    m = femagtools.mcv.MagnetizingCurve(mcv)
    result = m.writefile('TKS_NO_20', dir, fillfac=0.5)
    assert result == mcvPars[0]['name'] + '-50' + ext
    
    bi = [0.0, 0.04908391088247299, 0.09705951809883118,
          0.14508341252803802, 0.1947421133518219,
          0.24495112895965576, 0.2946207821369171,
          0.3442765474319458, 0.39504921436309814,
          0.44526469707489014, 0.4969058334827423,
          0.5512984395027161, 0.5992029309272766,
          0.6502805352210999, 0.7150968909263611,
          0.754536509513855, 0.7817665934562683,
          0.8623831868171692, 0.9199547171592712,
          0.9600813984870911, 0.9857978820800781,
          1.0011695623397827, 1.0120362043380737,
          1.0208477973937988]

    filename = os.path.join(dir, result)
    mcv = femagtools.mcv.read(filename)
    
    assert pytest.approx(bi) == mcv.get_results()['curve'][0]['bi']
    shutil.rmtree(dir)


    



"""

Geometry Parser for SVG files

"""
import logging
import re
import lxml.etree as ET
from .shape import Circle, Arc, Line, Element
import numpy as np

logger = logging.getLogger(__name__)

def get_center(r, p1, p2, sweep):
    """return center point coordinates of arc"""
    dp = p2-p1
    s = np.linalg.norm(dp)
    delta = np.arctan2(dp[1], dp[0])
    if s < 2*r:
        if sweep == 0:
            alfa = delta - np.arctan2(np.sqrt(r**2-s**2/4), s/2)
        else:
            alfa = delta + np.arctan2(np.sqrt(r**2-s**2/4), s/2)
    else:
        alfa = delta
    return p1[0] + r*np.cos(alfa), p1[1] + r*np.sin(alfa)


def get_angles(sweep, center, p1, p2):
    x1, y1 = (p1-center)
    x2, y2 = (p2-center)
    if sweep == 0:
        return np.arctan2(y2, x2), np.arctan2(y1, x1)
    return np.arctan2(y1, x1), np.arctan2(y2, x2)


def get_shapes(path):
    """return list of node elements (A, L)"""
    state = ''
    prevstate = ''
    p = []
    for s in [s for s in re.split('([AMLHV])|,|\\s+', path) if s]:
        if state == '':
            s = s.upper()
            if s in ('A','M','L','H','V'):
                state = s
                prevstate = s
            else:  # wild guess
                p.append(float(s))
                state = prevstate
        elif state == 'M':
            p.append(float(s))
            if len(p) == 2:
                p1 = np.array(p)
                p = []
                state = ''
        elif state == 'L':
            p.append(float(s))
            if len(p) == 2:
                p2 = np.array(p)
                logger.debug("Line %s -> %s",
                             p1, p2)
                yield Line(Element(start=p1, end=p2))
                p1 = p2.copy()
                p = []
                state = ''
        elif state == 'A':
            p.append(float(s))
            if len(p) == 7:
                sweep = int(p[-3])
                p2 = np.array(p[-2:])
                r = p[0]
                center = get_center(r, p1, p2, sweep)
                start, end = get_angles(sweep, center, p1, p2)
                logger.debug("Arc center %s r %f %f -> %f",
                             center, r, start, end)
                yield Arc(Element(center=center,
                                  radius=r,
                                  start_angle=start*180/np.pi,
                                  end_angle=end*180/np.pi))
                p1 = p2.copy()
                p = []
                state = ''
        elif state == 'H':
            logger.debug("h %s", s)
            p2 = np.array((float(s), 0))
            yield Line(Element(start=p1, end=p2))
            p1 = p2.copy()
            p = []
            state = ''
        elif state == 'V':
            logger.debug("V %s", s)
            p2 = np.array((0, float(s)))
            yield Line(Element(start=p1, end=p2))
            p1 = p2.copy()
            p = []
            state = ''
        else:
            raise ValueError(f"unsupported path {state}")


def svgshapes(svgfile):
    svg = ET.parse(svgfile)
    bcolor = re.compile('fill:([^;]+)')
    sr = 0
    for p in svg.findall(".//{http://www.w3.org/2000/svg}path"):
        m = bcolor.search(p.get('style'))
        if m:
            logger.debug("subregion %d: %s", sr, m.groups()[0])
        yield from get_shapes(p.get('d'))
        sr += 1
    for p in svg.findall(".//{http://www.w3.org/2000/svg}line"):
        yield Line(Element(start=[float(p.get('x1')), float(p.get('y1'))],
                           end=[float(p.get('x2')), float(p.get('y2'))]))
    for p in svg.findall(".//{http://www.w3.org/2000/svg}circle"):
        center = float(p.get('cx')), float(p.get('cy'))
        yield Circle(Element(center=center, radius=float(p.get('r'))))



#!/usr/bin/env python
#
import unittest
import femagtools

modelpars = dict(
    name="PM 130 L4",
    lfe=0.1,
    poles=4,
    stator=dict(
        yoke_diam=0.13,
        zeroangle=0.0,
        num_slots=12,
        inside_diam=0.07,
        mcvNameYoke="3",
        num_slots_gen=3,
        nodedist=1.5,
        rlength=1.0),

    magnet=dict(),
    windings=dict())


class ModelTest(unittest.TestCase):

    def test_getset(self):
        m = dict(modelpars)
        m['stator']['statorRotor3'] = dict(
            slot_h1=0.002,
            slot_h2=0.004,
            middle_line=0,
            tooth_width=0.009,
            wedge_width2=0.0,
            wedge_width1=0.0,
            slot_top_sh=0,
            slot_r2=0.002,
            slot_height=0.02,
            slot_r1=0.003,
            slot_width=0.003)
        model = femagtools.MachineModel(m)
        attr = 'stator.statorRotor3.slot_height'.split('.')
        self.assertEqual(model.get(attr), 20e-3)
        newvalue = 1
        model.set_value(attr, newvalue)
        self.assertEqual(model.get(attr), newvalue)

    def test_get_props(self):
        m = dict(modelpars)
        model = femagtools.MachineModel(m)
        self.assertEqual(model.props(),
                          {'name': 'PM130L4', 'poles': 4,
                           'external_rotor': False,
                           'stator': {'num_slots_gen':3, 'num_slots':12}})

    def test_magnet_material(self):
        m = dict(modelpars)
        m['magnet']['magnetSector'] = dict()
        m['magnet']['material'] = dict(name='MCV')
        model = femagtools.MachineModel(m)

        self.assertEqual(model.magnettype(), 'magnetSector')
        self.assertEqual(model.magnet['material'], {'name': 'MCV'})

    def test_modelname(self):
        modelname = 'PM-45_56-B'
        model = femagtools.MachineModel({'name': modelname})
        self.assertEqual(modelname, model.name)
        modelname = 'PM-45/?@56-B'
        model = femagtools.MachineModel({'name': modelname})
        self.assertEqual('PM-4556-B', model.name)

    def test_unchanged(self):
        m = dict(modelpars)
        model = femagtools.MachineModel(m)
        decision_vars = [{'name':   'current',
                          'steps':  2,
                          'bounds': [50.0, 100.0]},
                         {'name':   'windings.num_wires',
                          'steps':  3,
                          'bounds': [10.0, 20.0]}]
        matched_attrs = [d for d in decision_vars
                         if hasattr(model,
                                    d['name'].split('.')[0])]
        self.assertTrue(matched_attrs)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
#
import unittest
import sys
sys.path.insert(0, '../..')
from moo.problem import Problem


class ProblemTest(unittest.TestCase):
    def test_compare_fc(self):
        prob = Problem(0, 0, 2)
        f1 = (0, 0)
        f2 = (1, 1)
        c1 = []
        c2 = []
        self.assertTrue(prob.compare_fc(f1, c1, f2, c2))
        self.assertFalse(prob.compare_fc(f2, c2, f1, c1))
        self.assertFalse(prob.compare_fc(f2, c2, f2, c2))

if __name__ == '__main__':
  unittest.main()



import pytest
import pathlib
from femagtools import vtu


@pytest.fixture
def ts_data_dir():
    return pathlib.Path(__file__).with_name('data') / 'zzz_pm_model_ts_results_1/'

@pytest.fixture
def demag_data_dir():
    return pathlib.Path(__file__).with_name('data') / 'demag-vtu/'


def test_read(ts_data_dir):
    vtu_data = vtu.read(ts_data_dir / 'zzz_pm_model_ts_0000.vtu')
    assert vtu_data.field_data_names == [
        'time [s]', 'angle [rad]',
        'speed [rad/s]', 'torque [Nm]',
        'winding 1 current [A]', 'winding 1 voltage [V]',
        'winding 2 current [A]', 'winding 2 voltage [V]',
        'winding 3 current [A]', 'winding 3 voltage [V]', 'branch names']
    assert vtu_data.point_data_names == ['vector potential']
    assert vtu_data.cell_data_names == [
        'b', 'curd', 'demagnetization', 'current [A]', 'voltage [V]']

    b = vtu_data.get_data_vector('b', 1)
    assert b[0] == pytest.approx([-0.003114], abs=1e-5)
    assert b[1] == pytest.approx([-0.00313], abs=1e-5)
    assert b[2] == [0.0]

def test_demag(demag_data_dir):
    vtu_data = vtu.read(demag_data_dir / 'PM_130_L10_0000.vtu')
    keys = [7412, 7413, 7414, 7415, 7416]
    class Element:
        def __init__(self, key):
            self.key = key
    elements = [Element(k) for k in keys]
    expected = [241.5, 250.2, 262.3, 276.5, 390.5]
    actual = vtu_data.demag(elements)
    assert len(actual) == 1
    assert actual[0] == pytest.approx(expected, abs=0.1)
